@@ -74,6 +74,7 @@ class BaseAdapter(object):
     interpolation = '%s'
     sequence_support = False
     reserved_tables = []
+    quote_char = '"'
     
     def get_field_types(self):
         field_types = {
@@ -131,6 +132,7 @@ class SqliteAdapter(BaseAdapter):
         'ne': '!= %s', # watch yourself with this one
         'in': 'IN (%s)', # special-case to list q-marks
         'is': 'IS %s',
+        'isnull': 'IS NULL',
         'between': 'BETWEEN %s AND %s',
         'icontains': "LIKE %s ESCAPE '\\'", # surround param with %'s
         'contains': "GLOB %s", # surround param with *'s
@@ -166,6 +168,7 @@ class PostgresqlAdapter(BaseAdapter):
         'ne': '!= %s', # watch yourself with this one
         'in': 'IN (%s)', # special-case to list q-marks
         'is': 'IS %s',
+        'isnull': 'IS NULL',
         'between': 'BETWEEN %s AND %s',
         'icontains': 'ILIKE %s', # surround param with %'s
         'contains': 'LIKE %s', # surround param with *'s
@@ -210,12 +213,14 @@ class MySQLAdapter(BaseAdapter):
         'ne': '!= %s', # watch yourself with this one
         'in': 'IN (%s)', # special-case to list q-marks
         'is': 'IS %s',
+        'isnull': 'IS NULL',
         'between': 'BETWEEN %s AND %s',
         'icontains': 'LIKE %s', # surround param with %'s
         'contains': 'LIKE BINARY %s', # surround param with *'s
         'istartswith': 'LIKE %s',
         'startswith': 'LIKE BINARY %s',
     }
+    quote_char = '`'
 
     def connect(self, database, **kwargs):
         if not mysql:
@@ -321,6 +326,9 @@ class Database(object):
     
     def rows_affected(self, cursor):
         return self.adapter.rows_affected(cursor)
+
+    def quote_name(self, name):
+        return ''.join((self.adapter.quote_char, name, self.adapter.quote_char))
     
     def column_for_field(self, field):
         return self.column_for_field_type(field.get_db_field())
@@ -332,7 +340,10 @@ class Database(object):
             raise AttributeError('Unknown field type: "%s", valid types are: %s' % \
                 db_field_type, ', '.join(self.adapter.get_field_types().keys())
             )
-    
+
+    def field_sql(self, field):
+        return '%s %s' % (self.quote_name(field.name), field.render_field_template())
+
     def create_table(self, model_class, safe=False):
         if model_class._meta.pk_sequence and self.adapter.sequence_support:
             if not self.sequence_exists(model_class._meta.pk_sequence):
@@ -341,14 +352,18 @@ class Database(object):
         columns = []
 
         for field in model_class._meta.get_fields():
-            columns.append(field.to_sql())
+            columns.append(self.field_sql(field))
 
-        query = framing % (model_class._meta.db_table, ', '.join(columns))
+        table = self.quote_name(model_class._meta.db_table)
+        query = framing % (table, ', '.join(columns))
         
         self.execute(query)
     
     def create_index(self, model_class, field_name, unique=False):
-        framing = 'CREATE %(unique)s INDEX %(model)s_%(field)s ON %(model)s(%(field)s);'
+        framing = 'CREATE %(unique)s INDEX %(index)s ON %(table)s(%(field)s);'
+
+        db_table = model_class._meta.db_table
+        index_name = self.quote_name('%s_%s' % (db_table, field_name))
         
         if field_name not in model_class._meta.fields:
             raise AttributeError(
@@ -359,8 +374,9 @@ class Database(object):
         
         query = framing % {
             'unique': unique_expr,
-            'model': model_class._meta.db_table,
-            'field': field_name
+            'index': index_name,
+            'table': self.quote_name(db_table),
+            'field': self.quote_name(field_name),
         }
         
         self.execute(query)
@@ -370,15 +386,37 @@ class Database(object):
     
     def drop_table(self, model_class, fail_silently=False):
         framing = fail_silently and 'DROP TABLE IF EXISTS %s;' or 'DROP TABLE %s;'
-        self.execute(framing % model_class._meta.db_table)
+        self.execute(framing % self.quote_name(model_class._meta.db_table))
+    
+    def add_column_sql(self, model_class, field_name):
+        field = model_class._meta.fields[field_name]
+        return 'ALTER TABLE %s ADD COLUMN %s' % (
+            self.quote_name(model_class._meta.db_table),
+            self.field_sql(field),
+        )
+    
+    def rename_column_sql(self, model_class, field_name, new_name):
+        # this assumes that the field on the model points to the *old* fieldname
+        field = model_class._meta.fields[field_name]
+        return 'ALTER TABLE %s RENAME COLUMN %s TO %s' % (
+            self.quote_name(model_class._meta.db_table),
+            self.quote_name(field_name),
+            self.quote_name(new_name),
+        )
+    
+    def drop_column_sql(self, model_class, field_name):
+        return 'ALTER TABLE %s DROP COLUMN %s' % (
+            self.quote_name(model_class._meta.db_table),
+            self.quote_name(field_name),
+        )
     
     @require_sequence_support
     def create_sequence(self, sequence_name):
-        return self.execute('CREATE SEQUENCE %s;' % sequence_name)
+        return self.execute('CREATE SEQUENCE %s;' % self.quote_name(sequence_name))
     
     @require_sequence_support
     def drop_sequence(self, sequence_name):
-        return self.execute('DROP SEQUENCE %s;' % sequence_name)
+        return self.execute('DROP SEQUENCE %s;' % self.quote_name(sequence_name))
     
     def get_indexes_for_table(self, table):
         raise NotImplementedError
@@ -395,13 +433,19 @@ class SqliteDatabase(Database):
         super(SqliteDatabase, self).__init__(SqliteAdapter(), database, **connect_kwargs)
     
     def get_indexes_for_table(self, table):
-        res = self.execute('PRAGMA index_list(%s);' % table)
+        res = self.execute('PRAGMA index_list(%s);' % self.quote_name(table))
         rows = sorted([(r[1], r[2] == 1) for r in res.fetchall()])
         return rows
     
     def get_tables(self):
         res = self.execute('select name from sqlite_master where type="table" order by name')
         return [r[0] for r in res.fetchall()]
+    
+    def drop_column_sql(self, model_class, field_name):
+        raise NotImplementedError('Sqlite3 does not have direct support for dropping columns')
+    
+    def rename_column_sql(self, model_class, field_name, new_name):
+        raise NotImplementedError('Sqlite3 does not have direct support for renaming columns')
 
 
 class PostgresqlDatabase(Database):
@@ -443,9 +487,15 @@ class MySQLDatabase(Database):
     
     def create_foreign_key(self, model_class, field):
         framing = """ 
-            ALTER TABLE %(model)s ADD CONSTRAINT fk_%(model)s_%(to)s_%(field)s
+            ALTER TABLE %(table)s ADD CONSTRAINT %(constraint)s
             FOREIGN KEY (%(field)s) REFERENCES %(to)s(%(to_field)s)%(cascade)s;
         """
+        db_table = model_class._meta.db_table
+        constraint = 'fk_%s_%s_%s' % (
+            db_table,
+            field.to._meta.db_table,
+            field.name,
+        )
 
         if field.name not in model_class._meta.fields:
             raise AttributeError(
@@ -453,18 +503,28 @@ class MySQLDatabase(Database):
             )
  
         query = framing % {
-            'model': model_class._meta.db_table,
-            'field': field.name,
-            'to': field.to._meta.db_table,
-            'to_field': field.to._meta.pk_name,
+            'table': self.quote_name(db_table),
+            'constraint': self.quote_name(constraint),
+            'field': self.quote_name(field.name),
+            'to': self.quote_name(field.to._meta.db_table),
+            'to_field': self.quote_name(field.to._meta.pk_name),
             'cascade': ' ON DELETE CASCADE' if field.cascade else '',
         }
  
         self.execute(query)
         return super(MySQLDatabase, self).create_foreign_key(model_class, field)
-
+    
+    def rename_column_sql(self, model_class, field_name, new_name):
+        field = model_class._meta.fields[field_name]
+        return 'ALTER TABLE %s CHANGE COLUMN %s %s %s' % (
+            self.quote_name(model_class._meta.db_table),
+            self.quote_name(field_name),
+            self.quote_name(new_name),
+            field.render_field_template(),
+        )
+    
     def get_indexes_for_table(self, table):
-        res = self.execute('SHOW INDEXES IN %s;' % table)
+        res = self.execute('SHOW INDEXES IN %s;' % self.quote_name(table))
         rows = sorted([(r[2], r[1] == 0) for r in res.fetchall()])
         return rows
     
@@ -487,6 +547,9 @@ class QueryResultWrapper(object):
         self.column_meta = self.query_meta.get('columns')
         self.join_meta = self.query_meta.get('graph')
         
+        self.__ct = 0
+        self.__idx = 0
+        
         self._result_cache = []
         self._populated = False
     
@@ -561,16 +624,43 @@ class QueryResultWrapper(object):
         return inst
     
     def __iter__(self):
+        self.__idx = 0
+
         if not self._populated:
             return self
         else:
             return iter(self._result_cache)
     
+    def first(self):
+        try:
+            self.__idx = 0 # move to beginning of the list
+            inst = self.next()
+        except StopIteration:
+            inst = None
+        
+        self.__idx = 0
+        return inst
+
+    def fill_cache(self):
+        if not self._populated:
+            idx = self.__idx
+            self.__idx = self.__ct
+            for x in self:
+                pass
+            self.__idx = idx
+    
     def next(self):
+        if self.__idx < self.__ct:
+            inst = self._result_cache[self.__idx]
+            self.__idx += 1
+            return inst
+        
         row = self.cursor.fetchone()
         if row:
             instance = self.construct_instance(row)
             self._result_cache.append(instance)
+            self.__ct += 1
+            self.__idx += 1
             return instance
         else:
             self._populated = True
@@ -704,22 +794,17 @@ class F(object):
     def __init__(self, field, model=None):
         self.field = field
         self.model = model
-        self._op = None
-    
-    def sql(self):
-        if self._op is None:
-            return self.field
-        
-        return '(%s %s %s)' % (self.field, self._op, self.val)
+        self.op = None
+        self.value = None
 
     def __add__(self, rhs):
-        self._op = '+'
-        self.val = rhs
+        self.op = '+'
+        self.value = rhs
         return self
 
     def __sub__(self, rhs):
-        self._op = '-'
-        self.val = rhs
+        self.op = '-'
+        self.value = rhs
         return self
 
 
@@ -806,6 +891,9 @@ class BaseQuery(object):
     
     def clone(self):
         raise NotImplementedError
+
+    def qn(self, name):
+        return self.database.quote_name(name)
     
     def lookup_cast(self, lookup, value):
         return self.database.adapter.lookup_cast(lookup, value)
@@ -848,6 +936,9 @@ class BaseQuery(object):
                     raise ValueError('__is lookups only accept None')
                 operation = 'IS NULL'
                 lookup_value = []
+            elif op == 'isnull':
+                operation = 'IS NULL' if rhs else 'IS NOT NULL'
+                lookup_value = []
             elif isinstance(rhs, F):
                 lookup_value = rhs
                 operation = self.operations[op] # leave as "%s"
@@ -898,9 +989,15 @@ class BaseQuery(object):
         return len(self._joined_models) > 0 or self.force_alias
 
     def combine_field(self, alias, field_name):
+        quoted = self.qn(field_name)
         if alias:
-            return '%s.%s' % (alias, field_name)
-        return field_name
+            return '%s.%s' % (alias, quoted)
+        return quoted
+
+    def safe_combine(self, model, alias, col):
+        if col in model._meta.fields:
+            return self.combine_field(alias, col)
+        return col
     
     def follow_joins(self, current, alias_map, alias_required, alias_count, seen=None):
         computed = []
@@ -937,7 +1034,7 @@ class BaseQuery(object):
             computed.append(
                 '%s JOIN %s AS %s ON %s = %s' % (
                     join_type,
-                    model._meta.db_table,
+                    self.qn(model._meta.db_table),
                     alias_map[model],
                     self.combine_field(alias_map[from_model], left_field),
                     self.combine_field(alias_map[model], right_field),
@@ -1014,8 +1111,7 @@ class BaseQuery(object):
 
             if isinstance(value, F):
                 f_model = value.model or model
-                combined_value = self.combine_field(alias_map[f_model], value.sql())
-                operation = operation % combined_value
+                operation = operation % self.parse_f(value, f_model, alias_map)
             else:
                 query_data.append(value)
             
@@ -1031,6 +1127,13 @@ class BaseQuery(object):
             query = 'NOT %s' % query
         
         return query, query_data
+    
+    def parse_f(self, f_object, model, alias_map):
+        combined = self.combine_field(alias_map[model], f_object.field)
+        if f_object.op is not None:
+            combined = '(%s %s %s)' % (combined, f_object.op, f_object.value)
+
+        return combined
 
     def convert_subquery(self, subquery):
         subquery.query, orig_query = subquery.model._meta.pk_name, subquery.query
@@ -1211,14 +1314,15 @@ class SelectQuery(BaseQuery):
         return bool(curs.fetchone())
     
     def get(self, *args, **kwargs):
+        orig_ctx = self.query_context
+        self.query_context = self.model
+        query = self.where(*args, **kwargs).paginate(1, 1)
         try:
-            orig_ctx = self.query_context
-            self.query_context = self.model
-            obj = self.where(*args, **kwargs).paginate(1, 1).execute().next()
+            obj = query.execute().next()
             return obj
         except StopIteration:
             raise self.model.DoesNotExist('instance matching query does not exist:\nSQL: %s\nPARAMS: %s' % (
-                self.sql()
+                query.sql()
             ))
         finally:
             self.query_context = orig_ctx
@@ -1259,8 +1363,9 @@ class SelectQuery(BaseQuery):
             if model not in q:
                 continue
             
-            if q[model] == ['*']:
-                q[model] = model._meta.get_field_names()
+            if '*' in q[model]:
+                idx = q[model].index('*')
+                q[model] =  q[model][:idx] + model._meta.get_field_names() + q[model][idx+1:]
             
             for col in q[model]:
                 model_cols.append((model, col))
@@ -1269,15 +1374,15 @@ class SelectQuery(BaseQuery):
                     if len(col) == 3:
                         func, col, col_alias = col
                         aggregates.append('%s(%s) AS %s' % \
-                            (func, self.combine_field(alias, col), col_alias)
+                            (func, self.safe_combine(model, alias, col), col_alias)
                         )
                     elif len(col) == 2:
                         col, col_alias = col
                         columns.append('%s AS %s' % \
-                            (self.combine_field(alias, col), col_alias)
+                            (self.safe_combine(model, alias, col), col_alias)
                         )
                 else:
-                    columns.append(self.combine_field(alias, col))
+                    columns.append(self.safe_combine(model, alias, col))
         
         return ', '.join(columns + aggregates), model_cols
 
@@ -1286,19 +1391,23 @@ class SelectQuery(BaseQuery):
         joins, clauses, alias_map = self.compile_where()
         where, where_data = self.flatten_clauses(clauses)
         
-        table = self.model._meta.db_table
+        table = self.qn(self.model._meta.db_table)
 
         params = []
         group_by = []
+        use_aliases = self.use_aliases()
         
-        if self.use_aliases():
+        if use_aliases:
             table = '%s AS %s' % (table, alias_map[self.model])
-            for model, clause in self._group_by:
+
+        for model, clause in self._group_by:
+            if use_aliases:
                 alias = alias_map[model]
-                for field in clause:
-                    group_by.append(self.combine_field(alias, field))
-        else:
-            group_by = [c[1] for c in self._group_by]
+            else:
+                alias = ''
+
+            for field in clause:
+                group_by.append(self.combine_field(alias, field))
 
         parsed_query, model_cols = self.parse_select_query(alias_map)
         query_meta = {
@@ -1320,8 +1429,10 @@ class SelectQuery(BaseQuery):
         order_by = []
         for piece in self._order_by:
             model, field, ordering = piece
-            if self.use_aliases() and field in model._meta.fields:
-                field = '%s.%s' % (alias_map[model], field)
+            if field in model._meta.fields:
+                field = self.qn(field)
+                if self.use_aliases():
+                    field = '%s.%s' % (alias_map[model], field)
             order_by.append('%s %s' % (field, ordering))
         
         pieces = [select]
@@ -1409,7 +1520,7 @@ class UpdateQuery(BaseQuery):
 
         for k, v in set_statement.iteritems():
             if isinstance(v, F):
-                value = self.combine_field(alias, v.sql())
+                value = self.parse_f(v, v.model or self.model, alias_map)
             else:
                 params.append(v)
                 value = self.interpolation
@@ -1417,7 +1528,7 @@ class UpdateQuery(BaseQuery):
             update_params.append('%s=%s' % (self.combine_field(alias, k), value))
         
         update = 'UPDATE %s SET %s' % (
-            self.model._meta.db_table, ', '.join(update_params))
+            self.qn(self.model._meta.db_table), ', '.join(update_params))
         where = ' AND '.join(where)
         
         pieces = [update]
@@ -1451,7 +1562,7 @@ class DeleteQuery(BaseQuery):
 
         params = []
         
-        delete = 'DELETE FROM %s' % (self.model._meta.db_table)
+        delete = 'DELETE FROM %s' % (self.qn(self.model._meta.db_table))
         where = ' AND '.join(where)
         
         pieces = [delete]
@@ -1480,7 +1591,7 @@ class InsertQuery(BaseQuery):
         vals = []
         for k, v in self.insert_query.iteritems():
             field = self.model._meta.get_field_by_name(k)
-            cols.append(k)
+            cols.append(self.qn(k))
             vals.append(field.db_value(v))
         
         return cols, vals
@@ -1489,7 +1600,7 @@ class InsertQuery(BaseQuery):
         cols, vals = self.parse_insert()
         
         insert = 'INSERT INTO %s (%s) VALUES (%s)' % (
-            self.model._meta.db_table,
+            self.qn(self.model._meta.db_table),
             ','.join(cols),
             ','.join(self.interpolation for v in vals)
         )
@@ -1716,10 +1827,6 @@ class Field(object):
         self.attributes['column_type'] = col_type
         return self.field_template % self.attributes
     
-    def to_sql(self):
-        rendered = self.render_field_template()
-        return '%s %s' % (self.name, rendered)
-    
     def null_wrapper(self, value, default=None):
         if (self.null and value is None) or default is None:
             return value
